/* 
 * sada.c --
 *
 *	This file is an dynamically loadable TCL extension for audio control.
 *	This extension is Solaris/Sparc specific.
 *
 *      Subset of "audio.c", Copyright (c) 1997 by Sun Microsystems
 *	(c) 1997, 2007 Sun Microsystems, by Stephen Uhler
 *	See "license.txt" for license terms
 */

#include <sys/ioctl.h>
#include <sys/audioio.h>
#include <signal.h>
#include <stropts.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include "tcl.h"

#define dprintf	if(debug)fprintf
static int debug = 0;

extern int errno;

/*
 * Given a structure name and an element, compute the byte offset
 * to that element (stolen from Tk_Offset).
 */

#define OFFSET(type, field) ((int) ((char *) &((type *) 0)->field))
#define TCLVAR "AudioMonitorCallback"

/* types of audio dev structure elements */

#define T_INT 1
#define T_CHAR 2
#define T_DEVICE 4
#define T_BOOLEAN 8

struct fields {
    char *name;		/* the name of the option */
    int  format;	/* the format - how to print it */
    int offset;		/* structure offset */
};

/*
 * channel specific options
 */

static struct fields chan_options[] = {
    {"-active", T_CHAR|T_BOOLEAN, OFFSET(audio_prinfo_t,active)},
    {"-avail_ports", T_INT, OFFSET(audio_prinfo_t,avail_ports)},
    {"-balance", T_CHAR, OFFSET(audio_prinfo_t,balance)},
    {"-buffer_size", T_INT, OFFSET(audio_prinfo_t,buffer_size)},
    {"-channels", T_INT, OFFSET(audio_prinfo_t,channels)},
    {"-encoding", T_INT, OFFSET(audio_prinfo_t,encoding)},
    {"-eof", T_INT, OFFSET(audio_prinfo_t,eof)},
    {"-error", T_CHAR|T_BOOLEAN, OFFSET(audio_prinfo_t,error)},
    {"-gain", T_INT, OFFSET(audio_prinfo_t,gain)},
    {"-mod_ports", T_INT, OFFSET(audio_prinfo_t,mod_ports)},
    {"-pause", T_CHAR|T_BOOLEAN, OFFSET(audio_prinfo_t,pause)},
    {"-port", T_INT, OFFSET(audio_prinfo_t,port)},
    {"-precision", T_INT, OFFSET(audio_prinfo_t,precision)},
    {"-sample_rate", T_INT, OFFSET(audio_prinfo_t,sample_rate)},
    {"-samples", T_INT, OFFSET(audio_prinfo_t,samples)}, 
    {"-waiting", T_CHAR|T_BOOLEAN, OFFSET(audio_prinfo_t,waiting)},
    {(char *)NULL, 0, 0}
};

/*
 * Generic options.
 */

static struct fields gen_options[] = {
    {"-config", T_DEVICE, OFFSET(audio_device_t,config)},
    {"-monitor_gain", T_INT, OFFSET(audio_info_t,monitor_gain)},
    {"-name", T_DEVICE, OFFSET(audio_device_t,name)},
    {"-output_muted", T_CHAR|T_BOOLEAN, OFFSET(audio_info_t,output_muted)},
    {"-ref_count", T_CHAR, OFFSET(audio_info_t,ref_cnt)},
    {"-version", T_DEVICE, OFFSET(audio_device_t,version)},
    {(char *)NULL, 0, 0}
};


/*
 * arrays to hold indeces for ObjIndex function
 */

static char *gen_option_strings[sizeof gen_options / sizeof *gen_options];
static char *chan_option_strings[sizeof chan_options / sizeof *chan_options];

static Tcl_AsyncHandler async = NULL;	/* handler for sigPoll events */
static Tcl_Obj *varName;      /* TCL name for callback var */

/*
 * Declarations for functions defined in this file.
 */

static int AudioObjCmd _ANSI_ARGS_((ClientData dummy,
	Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));

static int AudioMonitor _ANSI_ARGS_((Tcl_Interp *interp, int objc,
	Tcl_Obj *CONST objv[], int fd));

static int notify _ANSI_ARGS_ ((ClientData clientData, Tcl_Interp *interp,
	int code));

static void poll(int sig);		/* wrong */

static int AudioPlayCmd _ANSI_ARGS_((Tcl_Interp *interp, int objc,
	Tcl_Obj *CONST objv[], int fd, struct audio_info info, int offset));

static int AudioGetValue _ANSI_ARGS_ ((Tcl_Interp *interp,
	struct fields *optionPtr, void *info));

static int AudioSetValue _ANSI_ARGS_ ((Tcl_Interp *interp,
	struct fields *optionPtr, void *info, void *value));

extern int Audio_Init _ANSI_ARGS_((Tcl_Interp *interp));


/*
 *----------------------------------------------------------------------
 *
 * Audio_Init --
 *
 *	This procedure initializes the Audio command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

extern int Audio_Init(interp)
    Tcl_Interp *interp;
{
    static struct fields *fieldsPtr;
    int i;

    /*
     * initialize obj-index arrays
     */

    for (i=0,fieldsPtr=chan_options; fieldsPtr->name;fieldsPtr++,i++) {
	chan_option_strings[i] = fieldsPtr->name;
    }
    chan_option_strings[i] = NULL;

    for (i=0,fieldsPtr=gen_options; fieldsPtr->name;fieldsPtr++,i++) {
	gen_option_strings[i] = fieldsPtr->name;
    }
    gen_option_strings[i] = NULL;

    Tcl_CreateObjCommand(interp, "audio", AudioObjCmd, NULL, NULL);
    varName = Tcl_NewStringObj(TCLVAR,-1);	/* this leaks alittle */
    return Tcl_PkgProvide(interp, "audio", "1.0");
}


/*
 *----------------------------------------------------------------------
 *
 * AudioCmd --
 *
 *	This function implements the Tcl "Audio" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
AudioObjCmd(dummy, interp, objc, objv)
    ClientData dummy;           /* Not used. */
    Tcl_Interp *interp;         /* Current interpreter. */
    int objc;                   /* Number of arguments. */
    Tcl_Obj *CONST objv[];      /* Argument objects. */
{
    int fd;			/* file description of audio device */
    int index = -1;		/* index into option list */
    int mode;			/* read/write mode for the channel */
    int len;			/* tcl string length */
    Tcl_Channel chan;		/* the audio device channel */
    struct audio_info info;	/* audio device info */
    struct audio_device device;	/* audio device identification info */
    struct audio_info info2;
    int i;
    int set = 0;

    static char *options[] = {
	 "record", "play", "monitor", "flush", "drain", "mark", (char *) NULL
     };

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv,
		"<audio_device_fd> monitor|record|play|flush|drain|mark ?-option value?...");
        return TCL_ERROR;
    }

    if (objc > 2 && *Tcl_GetStringFromObj(objv[2], &len) != '-' &&
	    Tcl_GetIndexFromObj(interp, objv[2], options, "option", 0, &index)
	    != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Find the audio device associated with this channel.
     */

    chan = Tcl_GetChannel(interp, Tcl_GetStringFromObj(objv[1], &len), &mode);
    if (chan == NULL) {
    	return TCL_ERROR;
    }
    if (Tcl_GetChannelHandle(chan, mode, (ClientData *)(&fd)) != TCL_OK) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't get handle for\"",
		Tcl_GetStringFromObj(objv[1], &len), "\": ",
		", invalid channel", (char *) NULL);
    	return TCL_ERROR;
    }
    if (ioctl(fd,AUDIO_GETDEV,&device) < 0) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't get status for\"",
		Tcl_GetStringFromObj(objv[1], &len), "\": ",
		" channel is not an audio device.", (char *) NULL);
    	return TCL_ERROR;
    }

    if (ioctl(fd,AUDIO_GETINFO,&info) < 0) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't get status for\"",
		Tcl_GetStringFromObj(objv[1], &len), "\": ",
		" confused audio device.", (char *) NULL);
    	return TCL_ERROR;
    }

    /*
     * Parse the arguments and handle the request.
     *
     * argument format:
     *  audio <fd>			Return the generic parameters
     *  audio <fd> record		Return current record settings
     *  audio <fd> play			Return current play settings
     *  audio <fd> -name		Return value of generic parameter
     *  audio <fd> record -name		Return value of record parameter
     *  audio <fd> play -name		Return value of record parameter
     *
     *  audio <fd> -name value ...	Set parameters...
     *  audio <fd> play -name value ...	
     *  audio <fd> record -name value ...
     *
     *  audio <fd> flush
     *  audio <fd> drain
     *  audio <fd> mark
     */


    /*
     * Do play or record specific stuff.
     */

    switch (index) {
	case 0:			/* record */
	    return  AudioPlayCmd(interp, objc, objv, fd, info,
			OFFSET(audio_info_t,record));
	    break;
	case 1:			/* play */
	    return  AudioPlayCmd(interp, objc, objv, fd, info,
			OFFSET(audio_info_t,play));
	    break;
	case 2:			/* set monitor callback */
	    return AudioMonitor(interp, objc, objv, fd);
	    break;
	case 3:			/* flush */
	    if (ioctl(fd, I_FLUSH, FLUSHRW) < 0) {
		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't flush audio", (char *)NULL);
		return TCL_ERROR;
	    }
	    return TCL_OK;
	    break;
	case 4:			/* drain */
	    if (ioctl(fd, AUDIO_DRAIN, NULL) < 0) {
		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't drain audio", (char *)NULL);
		return TCL_ERROR;
	    }
	    return TCL_OK;
	    break;
	case 5:			/* mark */
	    if (write(fd, &fd , 0) != 0) {
		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't write output mark", strerror(errno), (char *)NULL);
		return TCL_ERROR;
	    }
	    return TCL_OK;
	    break;
    }

    /*
     * Return all generic options
     */

    if (objc == 2) {
        struct fields *optionPtr;
	for(optionPtr=gen_options;optionPtr->name;optionPtr++) {
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
	    	" ", optionPtr->name, " ", NULL);
            void *tmp =optionPtr->format&T_DEVICE?(void *)&device:(void *)&info;
	    AudioGetValue(interp,optionPtr,tmp) ;
	}
	return TCL_OK;
    }

    /*
     *	Handle specific generic options 
     */

    if ((objc&1) && objc > 3) {
        Tcl_WrongNumArgs(interp, 1, objv,
		"Can't get and set values in 1 call.");
    	return TCL_ERROR;
    }

    /*
     * return a single generic object
     */

    index = -1;
    if (objc == 3 && Tcl_GetIndexFromObj(interp, objv[2],
	     gen_option_strings, "generic option", 0, &index) != TCL_OK) {
	return TCL_ERROR;
    }

    void *tmp =gen_options[index].format&T_DEVICE?(void *)&device:(void *)&info;
    if (index >= 0) {
	AudioGetValue(interp,&gen_options[index],tmp);
	return TCL_OK;
    }

    /*
     * Set groups of options
     */

    AUDIO_INITINFO(&info2);
    for (i = 2; i < objc; i +=2) {
	if (Tcl_GetIndexFromObj(interp, objv[i], gen_option_strings,
		"option", 0, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (AudioSetValue(interp,&gen_options[index],(void *)&info2,
		(void *) Tcl_GetStringFromObj(objv[i+1], &len)) != TCL_OK) {
	    return TCL_ERROR;
	}
	set++;
    }

    if (set && ioctl(fd,AUDIO_SETINFO,&info2) < 0) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't set parameters: ",strerror(errno), (char *)NULL);
    	return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 * set up the audio monitorring stuff 
 */

static int
AudioMonitor(interp, objc, objv, fd) 
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST objv[];
    int fd;
{
    Tcl_Obj *callback;
    callback = Tcl_ObjGetVar2(interp,varName,NULL, TCL_GLOBAL_ONLY);

    /*
     * return current callback object
     */

    dprintf(stderr,"Monitor fd=%d\n", fd);
    if (objc == 3) {
        if (callback != NULL) {
	    Tcl_SetObjResult(interp, callback);
	}
	return TCL_OK;
    }

    /*
     * set new callback object
     */

    if (objc == 4) {
    	int len;
	(void) Tcl_GetStringFromObj(objv[3], &len);
    	Tcl_ObjSetVar2(interp, varName, NULL, objv[3], TCL_GLOBAL_ONLY);
	if (async == NULL) {
	    async = Tcl_AsyncCreate(notify, (ClientData) interp);
	}
	if (len == 0) {
	    Tcl_UnsetVar(interp, TCLVAR,  TCL_GLOBAL_ONLY);
	    if (ioctl(fd,I_SETSIG,0) < 0) {
	    	perror("Can't clear I_SETSIG");
	    }
	    dprintf(stderr,"Removing up audio monitor - don't catch SIGPOLL\n");
	    if (sigrelse(SIGPOLL) != 0) {
	    	perror("Can't clear SIG_POLL");
	    }
	} else {
	    if (ioctl(fd,I_SETSIG,S_MSG) < 0) {
	    	perror("Can't set I_SETSIG");
	    }
	    if (sigset(SIGPOLL,poll) == SIG_ERR) {
	    	perror("Can't set SIG_POLL");
	    }
	    dprintf(stderr,"Setting up audio monitor - looking for SIGPOLL\n");
	}
	return TCL_OK;
    }
    Tcl_WrongNumArgs(interp, 1, objv, "monitor ?script?");
    return TCL_ERROR;
}

/*
 * run the notify code
 */

static int notify(clientData, interp, code) 
    ClientData clientData;
    Tcl_Interp *interp;
    int code;
{
    Tcl_Obj *callback, *saveResult;
    Tcl_Interp *myInterp;

    myInterp = (Tcl_Interp *) clientData;
    if (interp != NULL) {
	saveResult = Tcl_DuplicateObj(Tcl_GetObjResult(interp));
	Tcl_IncrRefCount(saveResult);
    }
    callback = Tcl_ObjGetVar2(myInterp,varName,NULL, TCL_GLOBAL_ONLY);
    dprintf(stderr,"Getting callback object, about to eval\n");
    if (callback == NULL) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(myInterp),
		"Can't find callback!!", (char *)NULL);
	Tcl_BackgroundError(myInterp);
    }
    if (Tcl_GlobalEvalObj(myInterp, callback) != TCL_OK) {
	Tcl_BackgroundError(myInterp);
    }
    if (interp != NULL) {
	Tcl_SetObjResult(interp, saveResult);
	Tcl_DecrRefCount(saveResult);
    }
    return (code);
}

static int
AudioPlayCmd(interp, objc, objv, fd, info, offset) 
    Tcl_Interp *interp;
    int objc;
    Tcl_Obj *CONST objv[];
    int fd;
    struct audio_info info;
    int offset;
{
    struct audio_info info2;
    int index;				/* index into option table */
    int i;
    struct fields *optionPtr;
    int len;
    int set = 0;

    if (objc >4 && (objc&1)==0) {
        Tcl_WrongNumArgs(interp, 1, objv, "Can't get and set values in 1 call.");
    	return TCL_ERROR;
    }

    /*
     * return all generic options.
     */

    if (objc == 3) {
	for(optionPtr=chan_options;optionPtr->name;optionPtr++) {
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
	    	" ", optionPtr->name, " ", NULL);
	    AudioGetValue(interp,optionPtr,(void *)&info+offset) ;
	}
	return TCL_OK;
    }

    /*
     * handle a single return option .
     */

    index = -1;
    if (objc == 4 && Tcl_GetIndexFromObj(interp, objv[3],
	     chan_option_strings, "option", 0, &index) != TCL_OK) {
	return TCL_ERROR;
    }

    if (index >= 0) {
	AudioGetValue(interp,&chan_options[index],(void *)&info + offset);
	return TCL_OK;
    }

    /*
     * Set groups of options
     */

    AUDIO_INITINFO(&info2);
    for (i = 3; i < objc; i +=2) {
	if (Tcl_GetIndexFromObj(interp, objv[i], chan_option_strings,
		"option", 0, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (AudioSetValue(interp,&chan_options[index],(void *)&info2+offset,
		(void *) Tcl_GetStringFromObj(objv[i+1], &len)) != TCL_OK) {
	    return TCL_ERROR;
	}
	set++;
    }

    if (set && ioctl(fd,AUDIO_SETINFO,&info2) < 0) {
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"couldn't set parameters: ",strerror(errno), (char *)NULL);
    	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 * Extract a result out of a structure.
 */

static int
AudioGetValue(interp,optionPtr,info) 
    Tcl_Interp *interp;		/* the interp to add the results to. */
    struct fields *optionPtr; /* which option to print */
    void *info;			/* pointer to the struct */
{  
    char buff[16];
    char *value = buff;
    unsigned int result;

    switch (optionPtr->format & 7) {
	case T_INT:
	    result = *(uint_t *)(((char *)(info)) + optionPtr->offset);
	    break;
	case T_CHAR:
	    result = *(((uchar_t *)(info)) + optionPtr->offset);
	    break;
	case T_DEVICE:
	    break;
	default:
	    fprintf(stderr,"Get value - Invalid type: %d\n",optionPtr->format);
	    /* abort(); */
	    return TCL_ERROR;
	    break;
    }
    if (optionPtr->format & T_DEVICE) {
	value = (((char *)(info)) + optionPtr->offset);
    } else if (optionPtr->format & T_BOOLEAN) {
	sprintf(buff,"%s", result ? "true" : "false");
    } else {
	sprintf(buff,"%d", result);
    }
    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
	    value, (char *) NULL);
    return TCL_OK;
}

/*
 * Set a result into a structure.
 */

static int
AudioSetValue(interp,optionPtr,info,value) 
    Tcl_Interp *interp;		/* the interp to add the results to. */
    struct fields *optionPtr;	/* which option to print */
    void *info;			/* pointer to the struct */
    void *value;		/* the value of pointer there-too */
{  
    int ok;
    int result = 0;
    if (optionPtr->format & T_BOOLEAN) {
	ok = Tcl_GetBoolean(interp, (char *) value, &result);
    } else {
	ok = Tcl_GetInt(interp, (char *) value, &result);
    }
    if (ok != TCL_OK) {
	return TCL_ERROR;
    }
    switch (optionPtr->format & 3) {
	case T_INT:
	    *(int *)(((char *)(info)) + optionPtr->offset) = result;
	    break;
	case T_CHAR:
	    *(((char *)(info)) + optionPtr->offset) = result;
	    break;
	default:
	    fprintf(stderr,"Set value - invalid type: %d\n",optionPtr->format);
	    /* abort(); */
	    return TCL_ERROR;
	    break;
    }
    return TCL_OK;
}

/*
 * got a signal, set the mark
 */

static
void poll(sig)
int sig;
{
    sighold(SIGPOLL);
    dprintf(stderr,"Got a sigpoll, calling AsyncMark\n");
    Tcl_AsyncMark(async);
    sigrelse(SIGPOLL);
}
